from __future__ import annotations

from pydantic import Field
from typing import Optional

from nsj_rest_lib2.compiler.edl_model.api_model import APIModel
from nsj_rest_lib2.compiler.edl_model.entity_model_root import EntityModelRoot


class EntityModel(EntityModelRoot):
    edl_version: Optional[str] = Field(default="1.0", description="Versão do EDL")
    version: Optional[str] = Field(
        default="1.0", description="Versão da entidade (padrão: 1.0)."
    )
    abstract: Optional[bool] = Field(
        default=False,
        description="Indica se a entidade é abstrata (padrão: False). Caso positivo, não gera código, mas pode ser usada na geração de código de outras entidades.",
    )
    api: APIModel = Field(
        ...,
        description="Definição da API REST associada ao modelo, com todos os seus endpoints.",
    )



from pydantic import BaseModel, Field
from typing import List, Optional, Literal

APIVerbs = Literal["GET", "POST", "PUT", "DELETE", "PATCH"]


class APIModel(BaseModel):
    resource: str = Field(
        ...,
        description="Nome do recurso REST (rota base dos endpoints; exemplo: 'clientes').",
    )
    expose: Optional[bool] = Field(
        default=True,
        description="Indica se a API deve ser exposta (padrão: True).",
    )
    verbs: Optional[List[APIVerbs]] = Field(
        default=["GET", "POST", "PUT", "DELETE", "PATCH"],
        description="Lista de verbos HTTP suportados pela API (padrão: todos).",
    )
    default_sort: Optional[List[str]] = Field(
        None,
        description="Lista de campos usados na ordenação padrão (padrão: se nada for fornecido, será usada, ao menos, a PK).",
    )



from typing import List, Optional
from pydantic import BaseModel, Field

from nsj_rest_lib2.compiler.edl_model.primitives import BasicTypes, PropertyType
from nsj_rest_lib2.compiler.edl_model.property_meta_model import DomainConfigModel


class TraitPropertyMetaModel(BaseModel):
    type: PropertyType = Field(..., description="Tipo da propriedade.")
    value: BasicTypes = Field(
        ..., description="Valor fixo da propriedade de condicionamento do trait."
    )
    domain_config: Optional[List[DomainConfigModel]] = Field(
        None,
        description="Lista de valores permitidos.",
    )


ExtendsPropertyMetaModel = TraitPropertyMetaModel



import ast
import re

from nsj_rest_lib2.compiler.compiler_structures import (
    IndexCompilerStructure,
    PropertiesCompilerStructure,
)
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import (
    BasicTypes,
    CardinalityTypes,
    PrimitiveTypes,
    REGEX_EXTERNAL_REF,
    REGEX_INTERNAL_REF,
)
from nsj_rest_lib2.compiler.edl_model.property_meta_model import PropertyMetaModel
from nsj_rest_lib2.compiler.edl_model.trait_property_meta_model import (
    TraitPropertyMetaModel,
)
from nsj_rest_lib2.compiler.model import RelationDependency
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import (
    compile_dto_class_name,
    compile_entity_class_name,
    compile_namespace_keys,
)
from nsj_rest_lib2.compiler.util.type_util import TypeUtil

# TODO pattern
# TODO lowercase
# TODO uppercase


class EDLPropertyCompiler:

    def compile(
        self,
        properties_structure: PropertiesCompilerStructure,
        map_unique_by_property: dict[str, IndexCompilerStructure],
        escopo: str,
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
        prefx_class_name: str,
    ) -> tuple[
        list[ast.stmt],
        list[ast.stmt],
        list[str],
        list[ast.stmt],
        list[tuple[str, str, str]],
        list[RelationDependency],
        list[tuple[str, BasicTypes]],
    ]:

        # TODO Criar opção de campo calculado?

        # Descobrindo os atributos marcados como PK (e recuperando a chave primária)
        # TODO Verificar se devemos manter essa verificação
        pk_keys = []
        for pkey in properties_structure.properties:
            prop = properties_structure.properties[pkey]

            if isinstance(prop.type, PrimitiveTypes):
                if prop.pk:
                    pk_keys.append(pkey)

        is_partial_extension = (
            isinstance(entity_model, EntityModel) and bool(entity_model.partial_of)
        )

        if not entity_model.mixin:
            if len(pk_keys) > 1:
                raise Exception(
                    f"Entidade '{entity_model.id}' possui mais de uma chave primária (ainda não suportado): {pk_keys}"
                )
            elif len(pk_keys) == 0 and not is_partial_extension:
                raise Exception(
                    f"Entidade '{entity_model.id}' não tem nenhuma chave primária (ainda não suportado)"
                )

        # pk_key = pk_keys[0]

        # Instanciando as listas de retorno
        ast_dto_attributes = []
        ast_entity_attributes = []
        props_pk = []
        aux_classes = []
        related_imports = []
        relations_dependencies = []
        fixed_filters = []

        if properties_structure.properties is None:
            return (ast_dto_attributes, ast_entity_attributes, props_pk, aux_classes)

        composed_properties = properties_structure.composed_properties or {}

        aggregator_class_names: dict[str, str] = {}
        aggregator_dto_attributes: dict[str, list[ast.stmt]] = {}
        aggregated_property_to_group: dict[str, str] = {}

        for composed_key, composed_list in composed_properties.items():
            if not composed_list:
                continue

            composed_class_name = (
                f"{CompilerStrUtil.to_pascal_case(escopo)}"
                f"{CompilerStrUtil.to_pascal_case(prefx_class_name)}"
                f"{CompilerStrUtil.to_pascal_case(entity_model.id)}"
                f"{CompilerStrUtil.to_pascal_case(composed_key)}AggregatorDTO"
            )

            aggregator_class_names[composed_key] = composed_class_name
            aggregator_dto_attributes[composed_key] = []

            for composed_property in composed_list:
                if composed_property in aggregated_property_to_group:
                    raise Exception(
                        f"Propriedade '{composed_property}' da entidade '{entity_model.id}' está associada a mais de um composed_property."
                    )

                if composed_property not in properties_structure.properties:
                    raise Exception(
                        f"Propriedade '{composed_property}' referenciada no composed_property '{composed_key}' não encontrada na entidade '{entity_model.id}'."
                    )

                aggregated_property_to_group[composed_property] = composed_key

        for pkey in properties_structure.properties:
            prop = properties_structure.properties[pkey]

            composed_key = aggregated_property_to_group.get(pkey)
            if composed_key:
                if prop.pk:
                    raise Exception(
                        f"Propriedade '{pkey}' não pode ser utilizada em composed_properties por ser chave primária."
                    )
                target_dto_attributes = aggregator_dto_attributes[composed_key]
            else:
                target_dto_attributes = ast_dto_attributes

            # DTO
            if isinstance(prop.type, PrimitiveTypes):
                # Tratando propriedade simples (não array, não object)
                self._compile_simple_property(
                    properties_structure,
                    map_unique_by_property,
                    escopo,
                    entity_model,
                    target_dto_attributes,
                    ast_entity_attributes,
                    props_pk,
                    aux_classes,
                    pkey,
                    prop,
                    prefx_class_name,
                )

            elif isinstance(prop.type, str):
                # Tratando propriedade de relacionamento
                external_match = re.match(REGEX_EXTERNAL_REF, prop.type)
                internal_match = re.match(REGEX_INTERNAL_REF, prop.type)

                if external_match:
                    # Resolvendo o id da entidade
                    related_entity_id = external_match.group(2)
                    related_entity_key = external_match.group(0)

                    self._compile_external_relation(
                        related_entity_id,
                        related_entity_key,
                        entity_model,
                        entity_models,
                        properties_structure,
                        target_dto_attributes,
                        ast_entity_attributes,
                        related_imports,
                        relations_dependencies,
                        pkey,
                        prop,
                    )

                elif internal_match:
                    related_entity_id = internal_match.group(1)

                    self._compile_internal_relation(
                        related_entity_id,
                        entity_model,
                        properties_structure,
                        target_dto_attributes,
                        ast_entity_attributes,
                        pkey,
                        prop,
                        prefx_class_name,
                    )
                else:
                    raise Exception(
                        f"Tipo da propriedade '{pkey}' não suportado: {prop.type}"
                    )

        for pkey in properties_structure.trait_properties:
            prop = properties_structure.trait_properties[pkey]

            self._compile_trait_extends_property(
                properties_structure,
                map_unique_by_property,
                pkey,
                prop,
                ast_dto_attributes,
                ast_entity_attributes,
                fixed_filters,
                escopo,
                entity_model,
                aux_classes,
                prefx_class_name,
            )

        for pkey in properties_structure.extends_properties:
            prop = properties_structure.extends_properties[pkey]

            self._compile_trait_extends_property(
                properties_structure,
                map_unique_by_property,
                pkey,
                prop,
                ast_dto_attributes,
                ast_entity_attributes,
                fixed_filters,
                escopo,
                entity_model,
                aux_classes,
                prefx_class_name,
            )

        for composed_key, class_name in aggregator_class_names.items():
            dto_attributes = aggregator_dto_attributes.get(composed_key, [])

            aux_classes.append(
                self._build_aggregator_class_ast(
                    class_name=class_name,
                    dto_attributes=dto_attributes,
                )
            )

            ast_dto_attributes.append(
                self._build_dto_aggregator_ast(
                    name=composed_key,
                    class_name=class_name,
                )
            )

        return (
            ast_dto_attributes,
            ast_entity_attributes,
            props_pk,
            aux_classes,
            related_imports,
            relations_dependencies,
            fixed_filters,
        )

    def _compile_trait_extends_property(
        self,
        properties_structure: PropertiesCompilerStructure,
        map_unique_by_property: dict[str, IndexCompilerStructure],
        pkey: str,
        prop: TraitPropertyMetaModel,
        ast_dto_attributes: list[ast.stmt],
        ast_entity_attributes: list[ast.stmt],
        fixed_filters: list[tuple[str, BasicTypes]],
        escopo: str,
        entity_model: EntityModelBase,
        aux_classes: list[ast.stmt],
        prefx_class_name: str,
    ):
        enum_class_name = None
        keywords = []

        if (
            properties_structure.main_properties
            and pkey in properties_structure.main_properties
        ):
            keywords.append(ast.keyword(arg="resume", value=ast.Constant(True)))

        if properties_structure.required and pkey in properties_structure.required:
            keywords.append(ast.keyword(arg="not_null", value=ast.Constant(True)))

        if (
            properties_structure.partition_data
            and pkey in properties_structure.partition_data
        ):
            keywords.append(ast.keyword(arg="partition_data", value=ast.Constant(True)))

        if pkey in map_unique_by_property:
            unique = map_unique_by_property[pkey].index_model
            keywords.append(
                ast.keyword(
                    arg="unique",
                    value=ast.Constant(unique.name),
                )
            )

        if (
            properties_structure.search_properties
            and pkey in properties_structure.search_properties
        ):
            keywords.append(ast.keyword(arg="search", value=ast.Constant(True)))
        else:
            keywords.append(ast.keyword(arg="search", value=ast.Constant(False)))

        if (
            properties_structure.metric_label
            and pkey in properties_structure.metric_label
        ):
            keywords.append(ast.keyword(arg="metric_label", value=ast.Constant(True)))

        if prop.type == PrimitiveTypes.CPF:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cpf",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.CNPJ:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cnpj",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.CPF_CNPJ:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cpf_or_cnpj",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.EMAIL:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_email",
                        ctx=ast.Load(),
                    ),
                )
            )

        # Trtando de uma definição de enum
        if prop.domain_config:
            result = self._compile_domain_config(
                pkey, prop, escopo, entity_model, prefx_class_name
            )
            if not result:
                raise Exception(f"Erro desconhecido ao compilar a propriedade {pkey}")

            enum_class_name, ast_enum_class = result
            aux_classes.append(ast_enum_class)

        # Resolvendo o nome da propriedade no Entity
        if (
            properties_structure.entity_properties
            and pkey in properties_structure.entity_properties
        ):
            entity_field_name = properties_structure.entity_properties[pkey].column
        else:
            entity_field_name = pkey

        # Escrevendo, se necessário, o alias para o nome da entity
        if entity_field_name != pkey:
            keywords.append(
                ast.keyword(
                    arg="entity_field",
                    value=ast.Constant(value=entity_field_name),
                )
            )

        # Instanciando o atributo AST
        if not isinstance(prop.type, PrimitiveTypes):
            raise Exception(
                f"Tipo da trait_property '{pkey}' não suportado: {prop.type} (deveria ser um Tipo Primitivo)"
            )

        # Instanciando o atributo AST
        if enum_class_name:
            prop_type = enum_class_name
        else:
            prop_type = TypeUtil.property_type_to_python_type(prop.type)

        ast_attr = ast.AnnAssign(
            target=ast.Name(id=CompilerStrUtil.to_snake_case(pkey), ctx=ast.Store()),
            annotation=ast.Name(
                id=prop_type,
                ctx=ast.Load(),
            ),
            value=ast.Call(
                func=ast.Name(id="DTOField", ctx=ast.Load()),
                args=[],
                keywords=keywords,
            ),
            simple=1,
        )

        ast_dto_attributes.append(ast_attr)

        # Entity
        ast_entity_attr = ast.AnnAssign(
            target=ast.Name(
                id=CompilerStrUtil.to_snake_case(entity_field_name),
                ctx=ast.Store(),
            ),
            annotation=ast.Name(
                id=TypeUtil.property_type_to_python_type(prop.type),
                ctx=ast.Load(),
            ),
            value=ast.Constant(value=None),
            simple=1,
        )

        ast_entity_attributes.append(ast_entity_attr)

        # Guardando como um fixed_filter
        # TODO Pensar em validações para esse value (se está de acordo com o tipo ou enum)
        fixed_filters.append((pkey, prop.value))

    def _build_aggregator_class_ast(
        self,
        class_name: str,
        dto_attributes: list[ast.stmt],
    ):
        body = dto_attributes if dto_attributes else [ast.Pass()]

        return ast.ClassDef(
            name=class_name,
            bases=[ast.Name(id="DTOBase", ctx=ast.Load())],
            keywords=[],
            decorator_list=[
                ast.Call(
                    func=ast.Name(id="DTO", ctx=ast.Load()),
                    args=[],
                    keywords=[],
                )
            ],
            body=body,
        )

    def _build_dto_aggregator_ast(
        self,
        name: str,
        class_name: str,
    ):
        return ast.AnnAssign(
            target=ast.Name(
                id=CompilerStrUtil.to_snake_case(name),
                ctx=ast.Store(),
            ),
            annotation=ast.Name(
                id=class_name,
                ctx=ast.Load(),
            ),
            value=ast.Call(
                func=ast.Name(id="DTOAggregator", ctx=ast.Load()),
                args=[ast.Name(id=class_name, ctx=ast.Load())],
                keywords=[],
            ),
            simple=1,
        )

    def _compile_external_relation(
        self,
        related_entity_id: str,
        related_entity_key: str,
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
        properties_structure: PropertiesCompilerStructure,
        ast_dto_attributes: list[ast.stmt],
        ast_entity_attributes: list[ast.stmt],
        related_imports: list[tuple[str, str, str]],
        relations_dependencies: list[RelationDependency],
        pkey: str,
        prop: PropertyMetaModel,
    ):
        # Resolvendo o nome das classes de DTO e Entity
        related_dto_class_name = compile_dto_class_name(related_entity_id)
        related_entity_class_name = compile_entity_class_name(related_entity_id)

        # Resolvendo o caminho do import
        related_entity = entity_models.get(related_entity_key)
        if not related_entity:
            raise Exception(
                f"Entidade '{entity_model.id}' possui uma referência externa para uma entidade inexistente: '{related_entity_key}', por meio da propriedade: '{pkey}'."
            )

        tenant = related_entity.tenant
        grupo_empresarial = related_entity.grupo_empresarial
        grupo_key, tenant_key, default_key = compile_namespace_keys(
            tenant, grupo_empresarial
        )

        if (
            tenant
            and tenant != 0
            and grupo_empresarial
            and grupo_empresarial != "00000000-0000-0000-0000-000000000000"
        ):
            related_import = grupo_key
        elif tenant and tenant != 0:
            related_import = tenant_key
        else:
            related_import = default_key

        related_imports.append(
            (
                related_import,
                related_dto_class_name,
                related_entity_class_name,
            )
        )

        # Gravando a dependência de relacionamento
        relation_dependency = RelationDependency()
        relation_dependency.entity_resource = related_entity.api.resource
        relation_dependency.entity_scope = related_entity.escopo
        relation_dependency.tenant = tenant
        relation_dependency.grupo_empresarial = grupo_empresarial
        relations_dependencies.append(relation_dependency)

        # Instanciando o ast
        if prop.cardinality == CardinalityTypes.C1_N:
            # Para relacionamentos 1_N
            self._build_ast_1_N(
                properties_structure,
                ast_dto_attributes,
                pkey,
                related_dto_class_name,
                related_entity_class_name,
                prop,
            )

        elif prop.cardinality == CardinalityTypes.C1_1:
            self._build_ast_1_1(
                properties_structure,
                ast_dto_attributes,
                ast_entity_attributes,
                pkey,
                related_dto_class_name,
                related_entity_class_name,
                prop,
            )

        elif prop.cardinality == CardinalityTypes.CN_N:
            # TODO
            pass
        else:
            raise Exception(
                f"Propriedade '{pkey}' da entidade '{entity_model.id}' possui cardinalidade inválida ou não suportada: {prop.cardinality}"
            )

    def _compile_internal_relation(
        self,
        related_entity_id: str,
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        ast_dto_attributes: list[ast.stmt],
        ast_entity_attributes: list[ast.stmt],
        pkey: str,
        prop: PropertyMetaModel,
        prefx_class_name: str,
    ):
        # Resolvendo o nome das classes de DTO e Entity
        related_dto_class_name = compile_dto_class_name(
            related_entity_id, f"{prefx_class_name}_{entity_model.id}"
        )
        related_entity_class_name = compile_entity_class_name(
            related_entity_id, f"{prefx_class_name}_{entity_model.id}"
        )

        # Instanciando o ast
        if prop.cardinality == CardinalityTypes.C1_N:
            # Para relacionamentos 1_N
            self._build_ast_1_N(
                properties_structure,
                ast_dto_attributes,
                pkey,
                related_dto_class_name,
                related_entity_class_name,
                prop,
            )

        elif prop.cardinality == CardinalityTypes.C1_1:
            self._build_ast_1_1(
                properties_structure,
                ast_dto_attributes,
                ast_entity_attributes,
                pkey,
                related_dto_class_name,
                related_entity_class_name,
                prop,
            )

        elif prop.cardinality == CardinalityTypes.CN_N:
            # TODO
            pass
        else:
            raise Exception(
                f"Propriedade '{pkey}' da entidade '{entity_model.id}' possui cardinalidade inválida ou não suportada: {prop.cardinality}"
            )

    def _build_ast_1_N(
        self,
        properties_structure: PropertiesCompilerStructure,
        ast_dto_attributes: list[ast.stmt],
        pkey: str,
        related_dto_class_name: str,
        related_entity_class_name: str,
        prop: PropertyMetaModel,
    ):
        # TODO Verificar uso da propriedade relation_key_field do Rest_lib_1

        # Propriedade do property descriptor
        keywords = [
            ast.keyword(
                arg="dto_type",
                value=ast.Name(id=related_dto_class_name, ctx=ast.Load()),
            ),
            ast.keyword(
                arg="entity_type",
                value=ast.Name(id=related_entity_class_name, ctx=ast.Load()),
            ),
        ]

        # Tratando das opções básicas do descritor de propriedade
        if properties_structure.required and pkey in properties_structure.required:
            keywords.append(ast.keyword(arg="not_null", value=ast.Constant(True)))

        if prop.max_length:
            keywords.append(ast.keyword(arg="max", value=ast.Constant(prop.max_length)))
        if prop.min_length:
            keywords.append(ast.keyword(arg="min", value=ast.Constant(prop.min_length)))

        if prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Name(prop.validator, ctx=ast.Load()),
                )
            )

        resume_fields = properties_structure.main_resume_fields.get(pkey)
        if resume_fields:
            keywords.append(
                ast.keyword(
                    arg="resume_fields",
                    value=ast.List(
                        elts=[ast.Constant(value=field) for field in resume_fields],
                        ctx=ast.Load(),
                    ),
                )
            )

        # Resolvendo a coluna usada no relacionamento
        if (
            not properties_structure.entity_properties
            or pkey not in properties_structure.entity_properties
            or not properties_structure.entity_properties[pkey].relation_column
        ):
            raise Exception(
                f"Propriedade '{pkey}' possui um relacionamento, mas nenhuma coluna de relacioanamento foi apontada na propriedade correspondente no repository."
            )

        relation_column_ref = properties_structure.entity_properties[
            pkey
        ].relation_column
        relation_column = str(relation_column_ref).split("/")[-1]

        keywords.append(
            ast.keyword(
                arg="related_entity_field",
                value=ast.Constant(value=relation_column),
            )
        )

        ast_attr = ast.AnnAssign(
            target=ast.Name(id=CompilerStrUtil.to_snake_case(pkey), ctx=ast.Store()),
            annotation=ast.Name(
                id="list",
                ctx=ast.Load(),
            ),
            value=ast.Call(
                func=ast.Name(id="DTOListField", ctx=ast.Load()),
                args=[],
                keywords=keywords,
            ),
            simple=1,
        )

        ast_dto_attributes.append(ast_attr)

    def _build_ast_1_1(
        self,
        properties_structure: PropertiesCompilerStructure,
        ast_dto_attributes: list[ast.stmt],
        ast_entity_attributes: list[ast.stmt],
        pkey: str,
        related_dto_class_name: str,
        related_entity_class_name: str,
        prop: PropertyMetaModel,
    ):
        # Propriedade do property descriptor
        keywords = [
            ast.keyword(
                arg="entity_type",
                value=ast.Name(id=related_entity_class_name, ctx=ast.Load()),
            ),
        ]

        # Tratando das opções básicas do descritor de propriedade
        if properties_structure.required and pkey in properties_structure.required:
            keywords.append(ast.keyword(arg="not_null", value=ast.Constant(True)))

        if (
            properties_structure.main_properties
            and pkey in properties_structure.main_properties
        ):
            keywords.append(ast.keyword(arg="resume", value=ast.Constant(True)))

        resume_fields = properties_structure.main_resume_fields.get(pkey)
        if resume_fields:
            keywords.append(
                ast.keyword(
                    arg="resume_fields",
                    value=ast.List(
                        elts=[ast.Constant(value=field) for field in resume_fields],
                        ctx=ast.Load(),
                    ),
                )
            )

        if prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Name(prop.validator, ctx=ast.Load()),
                )
            )

        # Resolvendo a coluna usada no relacionamento
        if (
            not properties_structure.entity_properties
            or pkey not in properties_structure.entity_properties
            or not properties_structure.entity_properties[pkey].relation_column
        ):
            raise Exception(
                f"Propriedade '{pkey}' possui um relacionamento, mas nenhuma coluna de relacioanamento foi apontada na propriedade correspondente no repository."
            )

        relation_column = str(
            properties_structure.entity_properties[pkey].relation_column
        )

        owner_relation = False
        if "/" in relation_column:
            owner_relation = True
            relation_column = relation_column.split("/")[-1]

        # TODO Verificar, porque desconfio que o apontamento para a coluna de relacionamento
        # para o caso do relacionamento OTHER, não é suportado pelo RestLib (acho que, quando
        # o dono é OTHER, o RestLib sempre aponta para a PK da entidade corrente).
        keywords.append(
            ast.keyword(
                arg="relation_field",
                value=ast.Constant(value=relation_column),
            )
        )

        if not owner_relation:
            keywords.append(
                ast.keyword(
                    arg="entity_relation_owner",
                    value=ast.Attribute(
                        value=ast.Name(id="EntityRelationOwner", ctx=ast.Load()),
                        attr="OTHER",
                        ctx=ast.Load(),
                    ),
                )
            )
        else:
            dto_property_name = f"relation_1_1_self_column_{relation_column}"
            # Adicionando propriedade, para o campo de relação, no DTO (quando for o dono da relação)
            ast_dto_attributes.append(
                self._build_dto_property_ast(
                    dto_property_name,
                    PrimitiveTypes.UUID,
                    keywords=[
                        ast.keyword(
                            arg="entity_field",
                            value=ast.Constant(value=relation_column),
                        ),
                    ],
                )
            )

            # Adicionando propriedade, para o campo de relação, no Entity (quando for o dono da relação)
            ast_entity_attributes.append(
                self._build_entity_property_ast(relation_column, PrimitiveTypes.UUID)
            )

        # Adicionando a propriedade em si do relacionamento, no DTO
        ast_attr = ast.AnnAssign(
            target=ast.Name(id=CompilerStrUtil.to_snake_case(pkey), ctx=ast.Store()),
            annotation=ast.Name(
                id=related_dto_class_name,
                ctx=ast.Load(),
            ),
            value=ast.Call(
                func=ast.Name(id="DTOObjectField", ctx=ast.Load()),
                args=[],
                keywords=keywords,
            ),
            simple=1,
        )

        ast_dto_attributes.append(ast_attr)

    def _build_dto_property_ast(
        self,
        name: str,
        type: PrimitiveTypes | str,
        keywords: list[ast.keyword] = [],
    ):
        if isinstance(type, PrimitiveTypes):
            type_str = TypeUtil.property_type_to_python_type(type)
        else:
            type_str = type

        return ast.AnnAssign(
            target=ast.Name(
                id=CompilerStrUtil.to_snake_case(name),
                ctx=ast.Store(),
            ),
            annotation=ast.Name(
                id=type_str,
                ctx=ast.Load(),
            ),
            value=ast.Call(
                func=ast.Name(id="DTOField", ctx=ast.Load()),
                args=[],
                keywords=keywords,
            ),
            simple=1,
        )

    def _build_entity_property_ast(
        self,
        name: str,
        type: PrimitiveTypes,
    ):
        return ast.AnnAssign(
            target=ast.Name(
                id=CompilerStrUtil.to_snake_case(name),
                ctx=ast.Store(),
            ),
            annotation=ast.Name(
                id=TypeUtil.property_type_to_python_type(type),
                ctx=ast.Load(),
            ),
            value=ast.Constant(value=None),
            simple=1,
        )

    def _compile_simple_property(
        self,
        properties_structure,
        map_unique_by_property,
        escopo,
        entity_model,
        ast_dto_attributes,
        ast_entity_attributes,
        props_pk,
        aux_classes,
        pkey,
        prop,
        prefx_class_name: str,
    ):
        enum_class_name = None
        keywords = []

        if prop.pk:
            keywords.append(ast.keyword(arg="pk", value=ast.Constant(True)))
            props_pk.append(pkey)

        if prop.key_alternative:
            keywords.append(ast.keyword(arg="candidate_key", value=ast.Constant(True)))

        if (
            properties_structure.main_properties
            and pkey in properties_structure.main_properties
        ):
            keywords.append(ast.keyword(arg="resume", value=ast.Constant(True)))

        if properties_structure.required and pkey in properties_structure.required:
            keywords.append(ast.keyword(arg="not_null", value=ast.Constant(True)))

        if (
            properties_structure.partition_data
            and pkey in properties_structure.partition_data
        ):
            keywords.append(ast.keyword(arg="partition_data", value=ast.Constant(True)))

        if pkey in map_unique_by_property:
            unique = map_unique_by_property[pkey].index_model
            keywords.append(
                ast.keyword(
                    arg="unique",
                    value=ast.Constant(unique.name),
                )
            )

        if (
            prop.default
        ):  # TODO Verificar esse modo de tratar valores default (principalmente expressões)
            keywords.append(
                ast.keyword(
                    arg="default_value",
                    value=ast.Name(str(prop.default), ctx=ast.Load()),
                )
            )

        if prop.trim:
            keywords.append(ast.keyword(arg="strip", value=ast.Constant(True)))

        max = None
        min = None
        if prop.type in [PrimitiveTypes.STRING, PrimitiveTypes.EMAIL]:
            if prop.max_length:
                max = prop.max_length
            if prop.min_length:
                min = prop.min_length
        elif prop.type in [PrimitiveTypes.INTEGER, PrimitiveTypes.NUMBER]:
            if prop.minimum:
                min = prop.minimum
            if prop.maximum:
                max = prop.maximum

        if max:
            keywords.append(ast.keyword(arg="max", value=ast.Constant(max)))
        if min:
            keywords.append(ast.keyword(arg="min", value=ast.Constant(min)))

        if (
            properties_structure.search_properties
            and pkey in properties_structure.search_properties
        ):
            keywords.append(ast.keyword(arg="search", value=ast.Constant(True)))
        else:
            keywords.append(ast.keyword(arg="search", value=ast.Constant(False)))

        if (
            properties_structure.metric_label
            and pkey in properties_structure.metric_label
        ):
            keywords.append(ast.keyword(arg="metric_label", value=ast.Constant(True)))

        if prop.type == PrimitiveTypes.CPF and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cpf",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.CNPJ and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cnpj",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.CPF_CNPJ and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cpf_or_cnpj",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.EMAIL and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_email",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Name(prop.validator, ctx=ast.Load()),
                )
            )

        if prop.immutable:
            keywords.append(ast.keyword(arg="read_only", value=ast.Constant(True)))

        if prop.on_save:
            keywords.append(
                ast.keyword(
                    arg="convert_to_entity",
                    value=ast.Name(prop.on_save, ctx=ast.Load()),
                )
            )

        if prop.on_retrieve:
            keywords.append(
                ast.keyword(
                    arg="convert_from_entity",
                    value=ast.Name(id=prop.on_retrieve, ctx=ast.Load()),
                )
            )

        if prop.domain_config:
            result = self._compile_domain_config(
                pkey, prop, escopo, entity_model, prefx_class_name
            )
            if not result:
                raise Exception(f"Erro desconhecido ao compilar a propriedade {pkey}")

            enum_class_name, ast_enum_class = result
            aux_classes.append(ast_enum_class)

        # Resolvendo o nome da propriedade no Entity
        if (
            properties_structure.entity_properties
            and pkey in properties_structure.entity_properties
        ):
            entity_field_name = properties_structure.entity_properties[pkey].column
        else:
            entity_field_name = pkey

        # Escrevendo, se necessário, o alias para o nome da entity
        if entity_field_name != pkey:
            keywords.append(
                ast.keyword(
                    arg="entity_field",
                    value=ast.Constant(value=entity_field_name),
                )
            )

        # Instanciando o atributo AST
        if enum_class_name:
            prop_type = enum_class_name
        else:
            prop_type = TypeUtil.property_type_to_python_type(prop.type)

        ast_attr = self._build_dto_property_ast(pkey, prop_type, keywords)
        ast_dto_attributes.append(ast_attr)

        # Entity
        ast_entity_attr = self._build_entity_property_ast(entity_field_name, prop.type)

        ast_entity_attributes.append(ast_entity_attr)

    def _compile_domain_config(
        self,
        pkey: str,
        prop: PropertyMetaModel | TraitPropertyMetaModel,
        escopo: str,
        entity_model: EntityModelBase,
        prefx_class_name: str,
    ) -> tuple[str, ast.stmt] | None:
        if not prop.domain_config:
            return None

        # Verificando se deveria usar o mapped_value
        use_mapped_value = False
        for value in prop.domain_config:
            if value.mapped_value:
                use_mapped_value = True
                break

        # Compilando as opções do enum
        ast_values = []
        for value in prop.domain_config:
            value_name = CompilerStrUtil.to_snake_case(value.value).upper()

            if use_mapped_value and value.mapped_value is None:
                raise Exception(
                    f"Propriedade '{pkey}' possui domain_config com value '{value.value}' mas sem mapped_value"
                )

            if value.mapped_value is not None:
                ast_value = ast.Assign(
                    targets=[ast.Name(id=value_name, ctx=ast.Store())],
                    value=ast.Tuple(
                        elts=[
                            ast.Constant(value=value.value),
                            ast.Constant(value=value.mapped_value),
                        ],
                        ctx=ast.Load(),
                    ),
                )
            else:
                ast_value = ast.Assign(
                    targets=[ast.Name(id=value_name, ctx=ast.Store())],
                    value=ast.Constant(value=value.value),
                )

            ast_values.append(ast_value)

        # Instanciando o atributo AST
        enum_class_name = f"{CompilerStrUtil.to_pascal_case(escopo)}{CompilerStrUtil.to_pascal_case(prefx_class_name)}{CompilerStrUtil.to_pascal_case(entity_model.id)}{CompilerStrUtil.to_pascal_case(pkey)}Enum"
        ast_enum_class = ast.ClassDef(
            name=enum_class_name,
            bases=[
                ast.Attribute(
                    value=ast.Name(id="enum", ctx=ast.Load()),
                    attr="Enum",
                    ctx=ast.Load(),
                )
            ],
            keywords=[],
            decorator_list=[],
            body=ast_values,
        )

        return enum_class_name, ast_enum_class



from __future__ import annotations

import uuid

from pydantic import BaseModel, Field
from typing import Dict, List, Optional

from nsj_rest_lib2.compiler.edl_model.property_meta_model import PropertyMetaModel
from nsj_rest_lib2.compiler.edl_model.repository_model import RepositoryModel
from nsj_rest_lib2.compiler.edl_model.trait_property_meta_model import (
    ExtendsPropertyMetaModel,
    TraitPropertyMetaModel,
)


class EntityModelBase(BaseModel):
    description: str = Field(..., description="Descrição da entidade.")
    id: str = Field(
        ...,
        description="Identificador único da entidade, dentro de um determinado escopo (podendo ser sobreescrito para um tenant ou grupo _empresarial).",
    )
    mixin: Optional[bool] = Field(
        False,
        description="Indica se o modelo é um mixin (bloco reutilizável, que não gera contém tabela própria no banco).",
    )
    extends: Optional[str] = Field(
        None, description="Identificador da entidade estendida por esta."
    )
    trait_from: Optional[str] = Field(
        None, description="Identificador da entidade que a trait estende."
    )
    partial_of: Optional[str] = Field(
        None,
        description="Identificador da entidade base estendida por esta extensão parcial (relacionamento 1x1 transparente).",
    )
    mixins: Optional[List[str]] = Field(
        None,
        description="Lista de mixins (blocos reutilizáveis) aplicados ao modelo.",
    )
    required: Optional[List[str]] = Field(
        None, description="Lista de campos obrigatórios no modelo."
    )
    partition_data: Optional[List[str]] = Field(
        None,
        description="Lista de propriedades da entidade, que serã usadas para particionamento dos dados no banco (sendo obrigatórias em todas as chamadas, inclusive como um tipo de filtro obrigatório).",
    )
    main_properties: Optional[List[str]] = Field(
        None,
        description="Lista de propriedades resumo do modelo (retornadas em qualquer chamada GET, mesmo que não pedidas por meio do parâmetro 'fields').",
    )
    search_properties: Optional[List[str]] = Field(
        None,
        description="Lista de propriedades do modelo que serão utilizadas nas chamadas de query do usuário (suportando também campos das entidades relacionadas). Sugere-se que a biblioteca crie automaticamente os índices para fulltext search.",
    )
    metric_label: Optional[List[str]] = Field(
        None,
        description="Lista de campos incluídos no agrupamento das métricas enviadas ao OpenTelemetry.",
    )
    trait_properties: Optional[Dict[str, TraitPropertyMetaModel]] = Field(
        None,
        description="Dicionário de propriedades condicionais da trait (isso é, especifica propriedades, e seus valores fixos, que serão usados como um tipo de filtro fixo usado como condição para que a propridade principal assum o papel daquele tipo de trait).",
    )
    extends_properties: Optional[Dict[str, ExtendsPropertyMetaModel]] = Field(
        None,
        description="Dicionário de propriedades condicionais associadas ao uso de extends (especifica propriedades e seus valores fixos, usados como filtros quando a entidade é uma especialização via extends).",
    )
    properties: Dict[str, PropertyMetaModel] = Field(
        ..., description="Dicionário de propriedades do modelo."
    )
    composed_properties: Optional[Dict[str, List[str]]] = Field(
        None,
        description="Dicionário que define grupos de propriedades compostas, permitindo agrupar propriedades existentes em DTOs agregados.",
    )
    components: Optional[Dict[str, "EntityModelBase"]] = Field(
        None,
        description="Entidades filhas, relacionadas por composição e definidas inline (evitando a criação de novos arquivos EDL para a definição de entidades componentes da entidade atual).",
    )
    repository: RepositoryModel = Field(
        ..., description="Configurações de mapeamento para o banco de dados."
    )

    # Propriedades de controle da compilação (não fazem parte do JSON de representação das entidades)
    tenant: int = Field(default=0, exclude=True)
    grupo_empresarial: uuid.UUID = Field(
        default=uuid.UUID("00000000-0000-0000-0000-000000000000"), exclude=True
    )


EntityModelBase.model_rebuild()



import re

from typing import Any

from nsj_rest_lib2.compiler.compiler_structures import (
    ComponentsCompilerStructure,
    IndexCompilerStructure,
    PropertiesCompilerStructure,
)
from nsj_rest_lib2.compiler.dto_compiler import DTOCompiler
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.entity_model_root import EntityModelRoot
from nsj_rest_lib2.compiler.edl_model.primitives import REGEX_EXTERNAL_REF
from nsj_rest_lib2.compiler.entity_compiler import EntityCompiler
from nsj_rest_lib2.compiler.model import CompilerResult, RelationDependency
from nsj_rest_lib2.compiler.property_compiler import EDLPropertyCompiler

from nsj_rest_lib2.compiler.util.type_naming_util import (
    compile_dto_class_name,
    compile_entity_class_name,
    compile_namespace_keys,
)

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel

from nsj_rest_lib2.settings import get_logger

# TODO GET e POST de relacionamentos 1X1
# TODO Revisar compilação de valores default (sensível a tipos)
# TODO Implementar suporte a conjuntos
# TODO Autenticação nas rotas
# TODO Atualizar o status da entidade pelo worker de compilação (e talvez parar uma compilação, quando se delete uma entidade)
# TODO Classes Abstratas
# TODO Partial Classes
# TODO Migrations
# TODO Criar imagem docker base para as aplicações, usando venv (para poderem atualizar o pydantic)

# TODO Suporte ao "min_items" dos relacionamentos no RestLib1


class EDLCompiler:
    def __init__(self) -> None:
        self._properties_compiler = EDLPropertyCompiler()
        self._dto_compiler = DTOCompiler()
        self._entity_compiler = EntityCompiler()

    def compile_models(
        self, entity_models: dict[str, EntityModel]
    ) -> list[CompilerResult]:

        compiler_results = []
        for entity_model_id in entity_models:
            entity_model = entity_models[entity_model_id]
            compiler_result = self._compile_model(entity_model, entity_models)
            if compiler_result:
                compiler_results.append(compiler_result)

        return compiler_results

    def compile_model_from_edl(
        self,
        edl_json: dict[str, Any],
        dependencies_edls: list[dict[str, Any]],
    ) -> CompilerResult | None:
        entity_model = EntityModel(**edl_json)

        entity_models = []
        for dependency_edl in dependencies_edls:
            if "mixin" in dependency_edl and dependency_edl["mixin"]:
                dependency_entity_model = EntityModelRoot(**dependency_edl)
            else:
                dependency_entity_model = EntityModel(**dependency_edl)
            entity_models.append(dependency_entity_model)

        return self.compile_model(entity_model, entity_models)

    def compile_model(
        self,
        entity_model: EntityModelBase,
        dependencies_models: list[tuple[str, EntityModelBase]],
    ) -> CompilerResult | None:
        entity_models = {}
        for dependency_entity_model in dependencies_models:
            complete_entity_id = dependency_entity_model[0]
            entity_models[complete_entity_id] = dependency_entity_model[1]

        return self._compile_model(entity_model, entity_models)

    def _compile_model(
        self,
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
        escopo: str | None = None,
        prefx_class_name: str = "",
    ) -> CompilerResult | None:
        if entity_model.mixin:
            return None

        if escopo is None and (isinstance(entity_model, EntityModel)):
            escopo = entity_model.escopo

        if not escopo:
            raise Exception(f"Escopo não definido para a entidade: {entity_model.id}.")

        # Tratando dos dados base para a extensão parcial
        partial_metadata: dict[str, str] | None = None
        partial_base_model: EntityModel | None = None
        if isinstance(entity_model, EntityModel) and entity_model.partial_of:
            if entity_model.partial_of not in entity_models:
                raise Exception(
                    f"Entidade base '{entity_model.partial_of}' não encontrada para a extensão parcial '{entity_model.id}'."
                )

            base_model_candidate = entity_models[entity_model.partial_of]
            if not isinstance(base_model_candidate, EntityModel):
                raise Exception(
                    f"Entidade base '{entity_model.partial_of}' da extensão parcial '{entity_model.id}' é inválida."
                )

            self._validate_partial_model(entity_model, base_model_candidate, entity_models)
            partial_metadata = self._build_partial_metadata(
                entity_model, base_model_candidate
            )
            partial_base_model = base_model_candidate

        # Criando um mapa de índices por nome de property
        # TODO Implementar tratamento dos índices de apoio às query (não de unicidade)
        map_indexes_by_property: dict[str, list[IndexCompilerStructure]] = {}
        map_unique_by_property: dict[str, IndexCompilerStructure] = {}
        self._make_unique_map_by_property(
            map_indexes_by_property, map_unique_by_property, entity_model, entity_models
        )

        # Criando uma cópia das coleções necessárias à compilação das properties
        # (a ideia é ser possível alterar as coleções sem afetar a entidade modelo,
        # o que será necessário para o tratamento de traits, etc - os quais serão
        # uma classe nova, resultado da união dessas propriedades).
        properties_structure = PropertiesCompilerStructure()
        self._make_properties_structures(
            properties_structure, entity_model, entity_models
        )

        # Criando a lista de atributos do DTO e da Entity; e recuperando as chaves primarias
        (
            ast_dto_attributes,
            ast_entity_attributes,
            props_pk,
            aux_classes,
            related_imports,
            relations_dependencies,
            fixed_filters,
        ) = self._properties_compiler.compile(
            properties_structure,
            map_unique_by_property,
            escopo,
            entity_model,
            entity_models,
            prefx_class_name,
        )

        # Adicionando os imports da extensão parcial
        if partial_metadata:
            related_imports.append(
                (
                    partial_metadata["module"],
                    partial_metadata["dto_class"],
                    partial_metadata["entity_class"],
                )
            )

        related_imports = self._deduplicate_related_imports(related_imports)

        # Gerando o buffer para os códigos de DTO e Entity
        dto_code = ""
        entity_code = ""
        relations_dependencies_complete = []

        # Carregando a estrutura de compilação dos components
        components_structure = ComponentsCompilerStructure()
        self._make_components_structures(
            components_structure, entity_model, entity_models
        )

        # Gerando o código das entidades filhas (components)
        for component_key in components_structure.components:
            component = components_structure.components[component_key]
            component_compiled = self._compile_model(
                component,
                entity_models,
                escopo,
                prefx_class_name=f"{prefx_class_name}_{entity_model.id}",
            )

            if not component_compiled:
                raise Exception(
                    f"Erro ao compilar o component '{component_key}' da entidade '{entity_model.id}'. Gerou saída None, como se fosse um mixin."
                )

            # Guardando o código gerado no buffer
            if component_compiled.dto_code:
                dto_code += component_compiled.dto_code + "\n\n"
            if component_compiled.entity_code:
                entity_code += component_compiled.entity_code + "\n\n"
            if component_compiled.relations_dependencies:
                relations_dependencies_complete.extend(
                    component_compiled.relations_dependencies
                )

        # Gerando o código do DTO
        dto_class_name, code_dto = self._dto_compiler.compile(
            entity_model,
            ast_dto_attributes,
            aux_classes,
            related_imports,
            fixed_filters,
            prefx_class_name,
            partial_metadata,
        )

        # Gerando o código da Entity
        entity_class_name, code_entity = self._entity_compiler.compile(
            entity_model,
            ast_entity_attributes,
            props_pk,
            prefx_class_name,
            partial_metadata,
        )

        # Extendendo os buffers com os códigos gerados
        dto_code += code_dto
        entity_code += code_entity

        # Adicionando as dependências das relações
        relations_dependencies_complete.extend(relations_dependencies)

        # Adicionando as dependências da extensão parcial
        if partial_metadata and partial_base_model:
            relation_dependency = RelationDependency()
            if not partial_base_model.api or not partial_base_model.api.resource:
                raise Exception(
                    f"Entidade base '{partial_base_model.id}' não possui configuração de API necessária para extensão parcial."
                )
            relation_dependency.entity_resource = partial_base_model.api.resource
            relation_dependency.entity_scope = partial_base_model.escopo
            relation_dependency.tenant = partial_base_model.tenant
            relation_dependency.grupo_empresarial = partial_base_model.grupo_empresarial
            relations_dependencies_complete.append(relation_dependency)

        # Construindo o resultado
        compiler_result = CompilerResult()
        compiler_result.entity_class_name = entity_class_name
        compiler_result.entity_code = entity_code
        compiler_result.dto_class_name = dto_class_name
        compiler_result.dto_code = dto_code
        compiler_result.relations_dependencies = relations_dependencies_complete

        # Compilando questões das APIs
        if isinstance(entity_model, EntityModel):
            compiler_result.api_expose = entity_model.api.expose
            compiler_result.api_resource = entity_model.api.resource
            compiler_result.api_verbs = entity_model.api.verbs

        get_logger().debug(f"código gerado para a entidade: {entity_model.id}")
        get_logger().debug("DTO Code:")
        get_logger().debug(f"\n{dto_code}")
        get_logger().debug("Entity Code:")
        get_logger().debug(f"\n{entity_code}")

        return compiler_result

    def _validate_partial_model(
        self,
        partial_model: EntityModel,
        base_model: EntityModel,
        entity_models: dict[str, EntityModel],
    ) -> None:
        base_properties_structure = PropertiesCompilerStructure()
        self._make_properties_structures(
            base_properties_structure,
            base_model,
            entity_models,
        )
        aggregated_base_properties = set(base_properties_structure.properties.keys())

        duplicated_properties = aggregated_base_properties.intersection(
            set(partial_model.properties.keys())
        )
        if duplicated_properties:
            raise Exception(
                f"Extensão parcial '{partial_model.id}' redefine propriedades da entidade base '{base_model.id}': {sorted(duplicated_properties)}."
            )

        lists_to_check = {
            "required": partial_model.required,
            "main_properties": partial_model.main_properties,
            "partition_data": partial_model.partition_data,
            "search_properties": partial_model.search_properties,
            "metric_label": partial_model.metric_label,
        }

        for list_name, values in lists_to_check.items():
            if not values:
                continue

            conflicts = [
                value
                for value in values
                if isinstance(value, str) and value in aggregated_base_properties
            ]
            if conflicts:
                raise Exception(
                    f"Extensão parcial '{partial_model.id}' utiliza propriedades da entidade base '{base_model.id}' em '{list_name}': {conflicts}."
                )

        link = partial_model.repository.link_to_base
        if not link:
            raise Exception(
                f"Extensão parcial '{partial_model.id}' requer configuração 'link_to_base' no bloco 'repository'."
            )

        if not link.base_property:
            raise Exception(
                f"Extensão parcial '{partial_model.id}' requer 'base_property' definido em 'link_to_base'."
            )

        if not link.column:
            raise Exception(
                f"Extensão parcial '{partial_model.id}' requer 'column' definido em 'link_to_base'."
            )

        if link.base_property not in aggregated_base_properties:
            raise Exception(
                f"'base_property' '{link.base_property}' não corresponde a uma propriedade da entidade base '{base_model.id}'."
            )

        # if not base_model.api or not base_model.api.resource:
        #     raise Exception(
        #         f"Entidade base '{base_model.id}' não possui configuração de API compatível com extensões parciais."
        #     )

    def _build_partial_metadata(
        self, partial_model: EntityModel, base_model: EntityModel
    ) -> dict[str, str]:
        link = partial_model.repository.link_to_base
        if link is None:
            raise Exception(
                f"Extensão parcial '{partial_model.id}' está sem configuração 'link_to_base'."
            )

        grupo_key, tenant_key, default_key = compile_namespace_keys(
            base_model.tenant, base_model.grupo_empresarial
        )
        namespace = self._resolve_namespace_key(
            base_model.tenant,
            base_model.grupo_empresarial,
            grupo_key,
            tenant_key,
            default_key,
        )

        return {
            "module": namespace,
            "dto_class": compile_dto_class_name(base_model.id),
            "entity_class": compile_entity_class_name(base_model.id),
            "relation_field": link.column,
            "related_entity_field": link.base_property,
        }

    def _resolve_namespace_key(
        self,
        tenant: str | int | None,
        grupo_empresarial,
        grupo_key: str,
        tenant_key: str,
        default_key: str,
    ) -> str:
        has_tenant = tenant not in (None, 0, "0")
        has_grupo = (
            grupo_empresarial
            and str(grupo_empresarial) != "00000000-0000-0000-0000-000000000000"
        )

        if has_tenant and has_grupo:
            return grupo_key
        if has_tenant:
            return tenant_key
        return default_key

    def _deduplicate_related_imports(
        self, related_imports: list[tuple[str, str, str]]
    ) -> list[tuple[str, str, str]]:
        deduped: list[tuple[str, str, str]] = []
        seen: set[tuple[str, str, str]] = set()

        for import_tuple in related_imports:
            if import_tuple in seen:
                continue
            seen.add(import_tuple)
            deduped.append(import_tuple)

        return deduped

    def _make_components_structures(
        self,
        components_structure: ComponentsCompilerStructure,
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
    ):
        if not entity_model:
            return

        # Populando com os components da superclasse (extends)
        if entity_model.extends:
            super_model = entity_models[entity_model.extends]

            self._make_components_structures(
                components_structure,
                super_model,
                entity_models,
            )

        # Populando com os components do trait
        if entity_model.trait_from:
            trait_model = entity_models[entity_model.trait_from]

            self._make_components_structures(
                components_structure,
                trait_model,
                entity_models,
            )

        # Populando com os components da entidade atual
        if entity_model.components:
            components_structure.components.update(entity_model.components)

    def _make_properties_structures(
        self,
        properties_structure: PropertiesCompilerStructure,
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
    ):
        if not entity_model:
            return

        # Populando com as propriedades dos mixins
        if entity_model.mixins:
            for mixin_id in entity_model.mixins:
                if mixin_id not in entity_models:
                    raise Exception(f"Mixin '{mixin_id}' não encontrado.")

                mixin_model = entity_models[mixin_id]
                self._make_properties_structures(
                    properties_structure,
                    mixin_model,
                    entity_models,
                )

        # Populando com as propriedades da superclasse (extends)
        if entity_model.extends:
            super_model = entity_models[entity_model.extends]

            self._make_properties_structures(
                properties_structure,
                super_model,
                entity_models,
            )

        # Populando com as propriedades do trait
        if entity_model.trait_from:
            trait_model = entity_models[entity_model.trait_from]

            self._make_properties_structures(
                properties_structure,
                trait_model,
                entity_models,
            )

        # Populando com as propriedades da entidade atual
        properties_structure.properties.update(entity_model.properties)
        if entity_model.main_properties:
            for main_property in entity_model.main_properties:
                if not isinstance(main_property, str):
                    continue

                if "/" in main_property:
                    path_parts = [
                        part.strip() for part in main_property.split("/") if part
                    ]
                    if len(path_parts) < 2 or not path_parts[0]:
                        raise Exception(
                            f"Propriedade resumo inválida '{main_property}' na entidade '{entity_model.id}'."
                        )

                    relation_name = path_parts[0]
                    resume_path_parts = path_parts[1:]
                    resume_field = ".".join(resume_path_parts)

                    if relation_name not in properties_structure.main_properties:
                        properties_structure.main_properties.append(relation_name)

                    resume_fields = properties_structure.main_resume_fields.setdefault(
                        relation_name, []
                    )

                    if resume_field and resume_field not in resume_fields:
                        resume_fields.append(resume_field)
                else:
                    if main_property not in properties_structure.main_properties:
                        properties_structure.main_properties.append(main_property)
        if entity_model.required:
            properties_structure.required.extend(entity_model.required)
        if entity_model.partition_data:
            properties_structure.partition_data.extend(entity_model.partition_data)
        if entity_model.search_properties:
            properties_structure.search_properties.extend(
                entity_model.search_properties
            )
        if entity_model.metric_label:
            properties_structure.metric_label.extend(entity_model.metric_label)

        if entity_model.trait_properties:
            properties_structure.trait_properties.update(entity_model.trait_properties)

        if entity_model.extends_properties:
            properties_structure.extends_properties.update(
                entity_model.extends_properties
            )

        if entity_model.composed_properties:
            properties_structure.composed_properties.update(
                entity_model.composed_properties
            )

        if entity_model.repository.properties:
            properties_structure.entity_properties.update(
                entity_model.repository.properties
            )

    def _make_unique_map_by_property(
        self,
        map_indexes_by_property: dict[str, list[IndexCompilerStructure]],
        map_unique_by_property: dict[str, IndexCompilerStructure],
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
        deep: int = 1,
    ):

        if not entity_model:
            return

        # Populando com as uniques da superclasse (extends)
        if entity_model.extends:
            super_model = entity_models[entity_model.extends]

            self._make_unique_map_by_property(
                map_indexes_by_property,
                map_unique_by_property,
                super_model,
                entity_models,
                deep=deep + 1,
            )

        # Populando com as uniques do trait
        if entity_model.trait_from:
            trait_model = entity_models[entity_model.trait_from]

            self._make_unique_map_by_property(
                map_indexes_by_property,
                map_unique_by_property,
                trait_model,
                entity_models,
                deep=deep + 1,
            )

        # Varrendo e organizando os índices
        if entity_model.repository.indexes:
            for index in entity_model.repository.indexes:
                for pkey in index.columns:
                    if index.unique:
                        if pkey in map_unique_by_property:
                            if deep > 1:
                                get_logger().warning(
                                    f"Propriedade '{pkey}' possui mais de um índice de unicidade (sendo um herdado). Por isso a replicação (herdada) será ignorada."
                                )
                                continue
                            else:
                                raise Exception(
                                    f"Propriedade '{pkey}' possui mais de um índice de unicidade."
                                )  # TODO Verificar esse modo de tratar erros

                        map_unique_by_property[pkey] = IndexCompilerStructure(
                            index, deep > 1
                        )
                    else:
                        list_index = map_indexes_by_property.setdefault(pkey, [])
                        list_index.append(IndexCompilerStructure(index, deep > 1))

    def list_dependencies(
        self, edl_json: dict[str, Any]
    ) -> tuple[list[str], EntityModelBase]:
        if edl_json.get("mixin", False):
            entity_model = EntityModelRoot(**edl_json)
        else:
            entity_model = EntityModel(**edl_json)

        return (self._list_dependencies(entity_model), entity_model)

    def _list_dependencies(self, entity_model: EntityModelBase) -> list[str]:
        entities: list[str] = []

        # Adicionando dependências por mixins
        if entity_model.mixins:
            entities.extend(entity_model.mixins)

        # Adicionando dependências por traits
        if entity_model.extends:
            entities.append(entity_model.extends)

        # Adicionando dependências por traits
        if entity_model.trait_from:
            entities.append(entity_model.trait_from)

        # Adicionando dependências por classes parciais
        if isinstance(entity_model, EntityModel) and entity_model.partial_of:
            entities.append(entity_model.partial_of)

        # Populando com as dependências de propriedades de relacionamento
        relations = self._list_dependencies_relations(entity_model)

        components_dependency_list = []
        if entity_model.components is not None:
            for component in entity_model.components:
                components_dependency_list.extend(
                    self._list_dependencies(entity_model.components[component])
                )

        relations.extend(components_dependency_list)

        entities.extend(relations)

        return entities

    def _list_dependencies_relations(self, entity_model) -> list[str]:
        entities = []

        # Relacionamento 1_N
        for pkey in entity_model.properties:
            prop = entity_model.properties[pkey]

            if isinstance(prop.type, str):
                external_match = re.match(REGEX_EXTERNAL_REF, prop.type)
                if external_match:
                    external_dependency = external_match.group(0)
                    entities.append(external_dependency)

        return entities


def get_files_from_directory(directory):
    files = []
    for file in os.listdir(directory):
        if file.endswith(".json") or file.endswith(".yml") or file.endswith(".yaml"):
            files.append(os.path.join(directory, file))
    return files


if __name__ == "__main__":
    import argparse
    import json
    import os
    import yaml

    parser = argparse.ArgumentParser(
        description="Compila arquivos EDL para classes Python"
    )
    parser.add_argument(
        "-d",
        "--directory",
        help="Diretório com arquivos .json para compilar",
        type=str,
        required=True,
    )
    args = parser.parse_args()

    files = get_files_from_directory(args.directory)

    entities = {}
    for file in files:
        with open(file, "r") as f:
            if file.endswith(".json"):
                edl = json.load(f)
            else:
                edl = yaml.safe_load(f)

        # Instanciando o objeto de modelo de entidade a partir do JSON,
        # e já realizando as validações básicas de tipo e estrutura.
        print(f"Validando arquivo: {file}")
        if edl.get("mixin", False):
            entity_model = EntityModelRoot(**edl)
        else:
            entity_model = EntityModel(**edl)

        complete_entity_id = f"{entity_model.escopo}/{entity_model.id}"
        entities[complete_entity_id] = entity_model

    compiler = EDLCompiler()
    compiler_results = compiler.compile_models(entities)

    with open("output_compilacao_local.py", "w") as f:
        for compiler_result in compiler_results:
            f.write("==========================================================\n")
            f.write(f"Entity: {compiler_result.entity_class_name}\n")
            f.write(f"{compiler_result.entity_code}\n")
            f.write("\n")
            f.write("==========================================================\n")
            f.write(f"DTO: {compiler_result.dto_class_name}\n")
            f.write(f"{compiler_result.dto_code}\n")
            f.write("\n")
            f.write("==========================================================\n")
            f.write(f"API Expose: {compiler_result.api_expose}\n")
            f.write(f"API Route Path: {compiler_result.api_resource}\n")
            f.write(f"API Verbs: {compiler_result.api_verbs}\n")
            f.write("==========================================================\n")
            f.write("\n")



import ast

import black

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import compile_entity_class_name


class EntityCompiler:
    def __init__(self):
        pass

    def compile(
        self,
        entity_model: EntityModelBase,
        ast_entity_attributes: list[ast.stmt],
        props_pk: list[str],
        prefix_class_name: str,
        partial_metadata: dict[str, str] | None,
    ) -> tuple[str, str]:
        # Imports
        imports = [
            # import datetime
            ast.Import(names=[ast.alias(name="datetime", asname=None)]),
            # import uuid
            ast.Import(names=[ast.alias(name="uuid", asname=None)]),
            # from nsj_rest_lib.entity.entity_base import EntityBase
            ast.ImportFrom(
                module="nsj_rest_lib.entity.entity_base",
                names=[ast.alias(name="EntityBase", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.decorator.entity import Entity
            ast.ImportFrom(
                module="nsj_rest_lib.decorator.entity",
                names=[ast.alias(name="Entity", asname=None)],
                level=0,
            ),
        ]

        if partial_metadata:
            imports.append(
                ast.ImportFrom(
                    module=f"dynamic.{partial_metadata['module']}",
                    names=[
                        ast.alias(name=partial_metadata["entity_class"], asname=None)
                    ],
                    level=0,
                )
            )

        # Entity
        if len(props_pk) > 1:
            raise Exception(
                f"Entidade '{entity_model.id}' possui mais de uma chave primária (ainda não suportado): {props_pk}"
            )

        if len(props_pk) == 0 and not partial_metadata:
            raise Exception(
                f"Entidade '{entity_model.id}' não possui nenhuma chave primária (ainda não suportado)."
            )

        default_order_props = []

        # Resolvendo o nome da coluna da chave primária
        if len(props_pk) == 1:
            key_field_property = props_pk[0]
        else:
            key_field_property = (
                partial_metadata["relation_field"] if partial_metadata else None
            )

        key_field = key_field_property
        if (
            len(props_pk) == 1
            and entity_model.repository.properties
            and key_field_property in entity_model.repository.properties
            and entity_model.repository.properties[key_field_property].column
        ):
            key_field = entity_model.repository.properties[key_field_property].column
        elif partial_metadata and partial_metadata.get("relation_field"):
            key_field = partial_metadata["relation_field"]

        if key_field is None:
            raise Exception(
                f"Não foi possível determinar a chave primária para a entidade '{entity_model.id}'."
            )

        if (
            isinstance(entity_model, EntityModel)
            and entity_model.api
            and entity_model.api.default_sort
        ):
            default_order_props = entity_model.api.default_sort

        default_order_fields = []
        for prop in default_order_props:
            if (
                entity_model.repository.properties
                and prop in entity_model.repository.properties
            ):
                field = entity_model.repository.properties[prop].column
            else:
                field = prop

            default_order_fields.append(CompilerStrUtil.to_snake_case(field))

        if CompilerStrUtil.to_snake_case(key_field) not in default_order_fields:
            default_order_fields.append(CompilerStrUtil.to_snake_case(key_field))

        class_name = compile_entity_class_name(entity_model.id, prefix_class_name)
        ast_class = ast.ClassDef(
            name=class_name,
            bases=[ast.Name(id="EntityBase", ctx=ast.Load())],
            keywords=[],
            decorator_list=[
                ast.Call(
                    func=ast.Name(id="Entity", ctx=ast.Load()),
                    args=[],
                    keywords=self._build_entity_decorator_keywords(
                        entity_model, key_field, default_order_fields, partial_metadata
                    ),
                )
            ],
            body=ast_entity_attributes,
        )

        # Definindo o módulo
        module = ast.Module(
            body=imports + [ast_class],
            type_ignores=[],
        )
        module = ast.fix_missing_locations(module)

        # Compilando o AST do DTO para o código Python
        code = ast.unparse(module)

        # Chamando o black para formatar o código Python do DTO
        code = black.format_str(code, mode=black.FileMode())

        return (class_name, code)

    def _build_entity_decorator_keywords(
        self,
        entity_model: EntityModelBase,
        key_field: str,
        default_order_fields: list[str],
        partial_metadata: dict[str, str] | None,
    ) -> list[ast.keyword]:
        keywords = [
            ast.keyword(
                arg="table_name",
                value=ast.Constant(value=entity_model.repository.map),
            ),
        ]

        if not partial_metadata:
            keywords.extend(
                [
                    ast.keyword(
                        arg="pk_field",
                        value=ast.Constant(
                            value=CompilerStrUtil.to_snake_case(key_field)
                        ),
                    ),
                    ast.keyword(
                        arg="default_order_fields",
                        value=ast.List(
                            elts=[
                                ast.Constant(value=field)
                                for field in default_order_fields
                            ],
                            ctx=ast.Load(),
                        ),
                    ),
                ]
            )

        if partial_metadata:
            keywords.insert(
                0,
                ast.keyword(
                    arg="partial_of",
                    value=ast.Name(id=partial_metadata["entity_class"], ctx=ast.Load()),
                ),
            )

        return keywords



import ast

import black

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import BasicTypes
from nsj_rest_lib2.compiler.util.type_naming_util import compile_dto_class_name


class DTOCompiler:
    def __init__(self):
        pass

    def compile(
        self,
        entity_model: EntityModelBase,
        ast_dto_attributes: list[ast.stmt],
        aux_classes: list[ast.stmt],
        related_imports: list[tuple[str, str, str]],
        fixed_filters: list[tuple[str, BasicTypes]],
        prefx_class_name: str,
        partial_metadata: dict[str, str] | None,
    ) -> tuple[str, str]:
        """
        Compila o código do DTO a partir do AST e retorna o código compilado.

        :param entity_model: Modelo de entidade
        :type entity_model: EntityModel

        :param ast_dto_attributes: Atributos do DTO
        :type ast_dto_attributes: list[ast.stmt]

        :param aux_classes: Classes auxiliares (enums, agregadores, etc.)
        :type aux_classes: list[ast.stmt]

        :return: Código compilado do DTO
        :rtype: str
        """
        # Criando o ast dos imports
        imports = [
            # import datetime
            ast.Import(names=[ast.alias(name="datetime", asname=None)]),
            # import enum
            ast.Import(names=[ast.alias(name="enum", asname=None)]),
            # import uuid
            ast.Import(names=[ast.alias(name="uuid", asname=None)]),
            # from nsj_rest_lib.decorator.dto import DTO
            ast.ImportFrom(
                module="nsj_rest_lib.decorator.dto",
                names=[ast.alias(name="DTO", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_field import DTOField
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_field",
                names=[ast.alias(name="DTOField", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_list_field import DTOField
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_list_field",
                names=[ast.alias(name="DTOListField", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_field_validators import DTOFieldValidators
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_field_validators",
                names=[ast.alias(name="DTOFieldValidators", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.dto.dto_base import DTOBase
            ast.ImportFrom(
                module="nsj_rest_lib.dto.dto_base",
                names=[ast.alias(name="DTOBase", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_left_join_field import EntityRelationOwner
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_left_join_field",
                names=[ast.alias(name="EntityRelationOwner", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_object_field import DTOObjectField
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_object_field",
                names=[ast.alias(name="DTOObjectField", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor import DTOAggregator
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor",
                names=[ast.alias(name="DTOAggregator", asname=None)],
                level=0,
            ),
        ]

        for import_ in related_imports:
            imports.append(
                ast.ImportFrom(
                    module=f"dynamic.{import_[0]}",
                    names=[ast.alias(name=import_[1])],
                    level=0,
                )
            )
            imports.append(
                ast.ImportFrom(
                    module=f"dynamic.{import_[0]}",
                    names=[ast.alias(name=import_[2])],
                    level=0,
                )
            )

        # Keywords para a extensão parcial
        decorator_keywords: list[ast.keyword] = []

        if partial_metadata:
            partial_dict_keys = [
                ast.Constant(value="dto"),
                ast.Constant(value="relation_field"),
            ]
            partial_dict_values = [
                ast.Name(id=partial_metadata["dto_class"], ctx=ast.Load()),
                ast.Constant(value=partial_metadata["relation_field"]),
            ]

            related_field = partial_metadata.get("related_entity_field")
            if related_field:
                partial_dict_keys.append(ast.Constant(value="related_entity_field"))
                partial_dict_values.append(ast.Constant(value=related_field))

            decorator_keywords.append(
                ast.keyword(
                    arg="partial_of",
                    value=ast.Dict(
                        keys=partial_dict_keys,
                        values=partial_dict_values,
                    ),
                )
            )

        # Keywords para tipos usados em fixed_filters
        if fixed_filters:
            decorator_keywords.append(
                ast.keyword(
                    arg="fixed_filters",
                    value=ast.Dict(
                        keys=[ast.Constant(value=item[0]) for item in fixed_filters],
                        values=[ast.Constant(value=item[1]) for item in fixed_filters],
                    ),
                )
            )

        # Criando o ast da classe
        class_name = compile_dto_class_name(entity_model.id, prefx_class_name)
        ast_class = ast.ClassDef(
            name=class_name,
            bases=[ast.Name(id="DTOBase", ctx=ast.Load())],
            keywords=[],
            decorator_list=[
                ast.Call(
                    func=ast.Name(id="DTO", ctx=ast.Load()),
                    args=[],
                    keywords=decorator_keywords,
                )
            ],
            body=ast_dto_attributes,
        )

        # Definindo o módulo
        module = ast.Module(
            body=imports + aux_classes + [ast_class],
            type_ignores=[],
        )
        module = ast.fix_missing_locations(module)

        # Compilando o AST do DTO para o código Python
        code = ast.unparse(module)

        # Chamando o black para formatar o código Python do DTO
        code = black.format_str(code, mode=black.FileMode())

        return (class_name, code)



from pydantic import BaseModel, Field
from typing import List, Optional

from nsj_rest_lib2.compiler.edl_model.column_meta_model import ColumnMetaModel
from nsj_rest_lib2.compiler.edl_model.index_model import IndexModel


class RepositoryLinkToBaseModel(BaseModel):
    base_property: str = Field(
        ...,
        description="Nome da propriedade na entidade base utilizada para o relacionamento 1x1.",
    )
    column: str = Field(
        ...,
        description="Nome da coluna (ou propriedade) na entidade parcial responsável por referenciar a entidade base.",
    )
    on_delete: Optional[str] = Field(
        None,
        description="Política de exclusão configurada no relacionamento com a entidade base.",
    )
    nullable: Optional[bool] = Field(
        False,
        description="Indica se o relacionamento com a entidade base pode ser nulo.",
    )


class RepositoryModel(BaseModel):
    map: str = Field(
        ..., description="Nome da tabela, no BD, para a qual a entidade é mapeada."
    )
    shared_table: Optional[bool] = Field(
        default=False,
        description="Indica se a tabela é compartilhada entre múltiplas entidades (padrão: False).",
    )
    table_owner: Optional[bool] = Field(
        None, description="Indica que essa entidade é dona da tabela (schema)."
    )  # TODO Validar explicação
    properties: Optional[dict[str, ColumnMetaModel]] = Field(
        None, description="Dicionário de colunas da entidade."
    )
    indexes: Optional[List[IndexModel]] = Field(
        None, description="Lista de índices de banco de dados, associados à entidade."
    )
    link_to_base: Optional[RepositoryLinkToBaseModel] = Field(
        None,
        description="Configuração de relacionamento 1x1 com a entidade base (extensões parciais).",
    )



from pydantic import BaseModel, Field, model_validator
from typing import List, Optional

from nsj_rest_lib2.compiler.edl_model.primitives import BasicTypes, DefaultTypes
from nsj_rest_lib2.compiler.edl_model.primitives import CardinalityTypes, PropertyType


class DomainConfigModel(BaseModel):
    value: str = Field(..., description="Valor do enum.")
    mapped_value: Optional[BasicTypes] = Field(
        None, description="Valor mapeado do enum (gravado no BD)."
    )


class PropertyMetaModel(BaseModel):
    type: PropertyType = Field(..., description="Tipo da propriedade.")
    label: Optional[str] = Field(None, description="Rótulo da propriedade.")
    description: Optional[str] = Field(None, description="Descrição da propriedade.")
    default: Optional[DefaultTypes] = Field(
        None,
        description="Valor padrão da propriedade (caso não fornecido numa chamada de persistência). Pode indicar um valor, ou uma expressão (ex.: 'now()' para data/hora atual).",
    )
    pk: Optional[bool] = Field(
        default=False, description="Indica se a propriedade é parte da chave primária."
    )
    key_alternative: Optional[bool] = Field(
        default=False,
        description="Indica se a propriedade é parte de uma chave alternativa (chave natural).",
    )
    cardinality: Optional[CardinalityTypes] = Field(
        None,
        description="Cardinalidade do relacionamento (válido para propriedades do tipo 'array' ou 'object').",
    )
    # Validação
    max_length: Optional[int] = Field(
        None, description="Comprimento máximo (para strings)."
    )

    ## TODO Sugestões de validação
    min_length: Optional[int] = Field(
        None, description="Comprimento mínimo (para strings)."
    )
    pattern: Optional[str] = Field(
        None,
        description="Expressão regular que o valor da propriedade deve obedecer (para strings).",
    )
    minimum: Optional[float] = Field(
        None, description="Valor mínimo permitido (para números)."
    )
    maximum: Optional[float] = Field(
        None, description="Valor máximo permitido (para números)."
    )
    validator: Optional[str] = Field(
        None,
        description="Caminho completo para uma função de validação e alteração personalizada (ex.: 'module.validator'). Sendo que as funções dinâmicas sempre estarão no modulo XPTO.",
    )  # TODO Definir padrão de carregamento do módulo de funções customizadas
    domain_config: Optional[List[DomainConfigModel]] = Field(
        None,
        description="Lista de valores permitidos.",
    )
    immutable: Optional[bool] = Field(
        default=False,
        description="Indica se a propriedade é imutável (não pode ser alterada após a criação).",
    )
    trim: Optional[bool] = Field(
        default=False,
        description="Indica se espaços em branco devem ser removidos do início e fim (para strings).",
    )
    lowercase: Optional[bool] = Field(
        default=False,
        description="Indica se o valor deve ser convertido para minúsculas (para strings).",
    )
    uppercase: Optional[bool] = Field(
        default=False,
        description="Indica se o valor deve ser convertido para maiúsculas (para strings).",
    )

    on_save: Optional[str] = Field(
        None,
        description="Função de conversão para o tipo da entidade (ex.: converter string para uuid).",
    )
    on_retrieve: Optional[str] = Field(
        None,
        description="Função de conversão do tipo da entidade para o tipo da API (ex.: converter uuid para string).",
    )

    @model_validator(mode="after")
    def check_cardinality_required(self):
        if isinstance(self.type, str) and self.cardinality is None:
            raise ValueError(
                "The property 'cardinality' is required when type point to other entity."
            )
        return self



from pydantic import BaseModel, Field, model_validator
from typing import Optional


class ColumnMetaModel(BaseModel):
    column: Optional[str] = Field(None, description="Nome da coluna no banco de dados.")
    relation_column: Optional[str] = Field(
        None,
        description="Nome da coluna usada para relacionamento com outra entidade (para um relacionamento 1_N, indica a coluna da entidade relacionada, que aponta para a PK da entidade corrente).",
    )

    @model_validator(mode="after")
    def check_columns(self):
        if self.column is None and self.relation_column is None:
            raise ValueError("column or relation_column must be provided")
        return self



from pydantic import Field

from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase


class EntityModelRoot(EntityModelBase):
    escopo: str = Field(..., description="Escopo do EDL (define a aplicação).")
